#!/usr/bin/python
# -*- coding: utf8 -*-
from subprocess import Popen, PIPE
from cStringIO import StringIO
import re
from IPy import IP
from checkdesc.descmodels import (Desc, Firewall, Network, Interface, Address,
    DirectConnection, InternetConnection, RoutedConnection)
from checkdesc.descxml import dump
from sys import exit, stderr

# Commands to get informations about network interfaces
IPADDR = ("/sbin/ip", "addr", "list")
IFCONFIG = ("/sbin/ifconfig", "-a")
ROUTE = ("/sbin/route", "-n")

# Default values in desc.xml
FIREWALL_QUEUE = 0
FIREWALL_NAME = "firewall"

# Special IPs
LOOPBACK = IP('127.0.0.0/8')
INTERNET = IP('0.0.0.0/0')

def searchFirst(condition, data):
    for item in data:
        if condition(item):
            return item
    raise KeyError('No item matching condition')

class TextParser:
    """
    Very basic plain text parser useful to read one line after the other.

    It calls a different function for each line, and each function returns
    next function to be called for next line.

    Interresting methods and attributes:
    - line_number is the current line number of input file (starting at 1)
    - reset(): function called when parser is created
    - stop(): function called when the parser is done
    - parserError(): raise an exception with reason and line number
    """
    def __init__(self, input, first_parser, filename=None):
        """
        Parse input file object, first_parser is the first function
        used to parse the file content.
        """
        self.input = input
        self.line_number = 0
        self.filename = filename
        self.first_parser = first_parser
        self.reset()
        self.runParser()

    def _formatError(self, error, message):
        if self.filename:
            where = "in %s, line %s" % (self.filename, self.line_number)
        else:
            where = "at line %s" % self.line_number
        return "%s %s: %s" % (error, where, message)

    def parserWarning(self, message):
        print >>stderr, self._formatError("Warning", message)

    def parserError(self, message):
        raise SyntaxError(self._formatError("Error", message))

    def reset(self):
        pass

    def stop(self):
        pass

    def runParser(self):
        parser = self.first_parser
        while True:
            line = self.input.readline()
            if len(line) == 0:
                break
            line = line.rstrip()
            self.line_number += 1
            new_parser = parser(line)
            if new_parser:
                parser = new_parser
        self.stop()

class IpAddrParser(TextParser):
    # eth0      Lien encap:Ethernet  HWaddr 00:15:C5:AA:35:38
    # lo        Lien encap:Boucle locale
    HEADER_REGEX = re.compile(r"^[0-9]+: ([a-z0-9]+): ")

    # link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
    LINK_REGEX = re.compile(r"^\s+link/([^ ]+)")

    # inet adr:192.168.33.170  Bcast:192.168.33.255  Masque:255.255.255.128
    # inet adr:127.0.0.1  Masque:255.0.0.0
    INET_REGEX = re.compile(r"^\s+inet ([0-9.]+(?:/[0-9]{1,2})?) ")

    def __init__(self, input, filename=None):
        self.interface = None
        self.interfaces = []
        self.interfaces_dict = {}
        TextParser.__init__(self, input, self.parseHeader, filename)

    def saveInterface(self):
        if self.interface is None:
            return
        name = self.interface['name']
        if 'addr' not in self.interface:
            self.parserWarning(
                "Unable to find IPv4 address of interface %s (skip this interface)"
                % name)
        elif name not in self.interfaces_dict:
            self.interfaces_dict[name] = self.interface
            self.interfaces.append(self.interface)
        self.interface = None

    def parseHeader(self, line):
        match = self.HEADER_REGEX.search(line)
        if match:
            self.saveInterface()
            name = match.group(1)
            try:
                self.interface = self.interfaces_dict[name]
            except KeyError:
                self.interface = {}
                self.interface['name'] = name
            return self.parseLink
        return self.parseHeader

    def parseLink(self, line):
        match = self.LINK_REGEX.search(line)
        if not match:
            self.parserError("Unable to parse link type: %r" % line)
        self.interface['link'] = match.group(1)
        return self.parseParameters

    def parseParameters(self, line):
        if self.HEADER_REGEX.search(line):
            return self.parseHeader(line)

        match = self.INET_REGEX.search(line)
        if match:
            addr = match.group(1)
            if '/' not in addr:
                addr += '/32'
            try:
                self.interface['addr'].append(addr)
            except KeyError:
                self.interface['addr'] = [addr]
        return self.parseParameters

    def stop(self):
        self.saveInterface()

class IfconfigParser(TextParser):
    # eth0      Lien encap:Ethernet  HWaddr 00:15:C5:AA:35:38
    # lo        Lien encap:Boucle locale
    HEADER_REGEX = re.compile(r"^([a-z0-9]+)(:[0-9]+)?\s+")

    # inet adr:192.168.33.170  Bcast:192.168.33.255  Mask:255.255.255.128
    # inet adr:127.0.0.1  Masque:255.0.0.0
    INET_REGEX = re.compile(r"^\s+inet [^: ]+:([^ ]+) +(?:[^: ]+:[^ ]+ +){0,2}[^: ]+:([0-9.]+)$")

    def __init__(self, input, filename=None):
        self.interface = None
        self.interfaces = []
        self.interfaces_dict = {}
        TextParser.__init__(self, input, self.parseHeader, filename)

    def saveInterface(self):
        if self.interface is None:
            return
        name = self.interface['name']
        if 'addr' not in self.interface:
            self.parserWarning(
                "Unable to find IPv4 address of interface %s (skip this interface)"
                % name)
        elif name.endswith(":avah"):
            self.parserWarning(
                "Skip \"avahi\" interace: %s" % name)
        elif name not in self.interfaces_dict:
            self.interfaces_dict[name] = self.interface
            self.interfaces.append(self.interface)
        self.interface = None

    def parseHeader(self, line):
        match = self.HEADER_REGEX.search(line)
        if match:
            self.saveInterface()
            name = match.group(1)
            try:
                self.interface = self.interfaces_dict[name]
            except KeyError:
                self.interface = {}
                self.interface['name'] = name
            return self.parseParameters
        return self.parseHeader

    def parseParameters(self, line):
        if not line:
            return self.parseHeader
        match = self.INET_REGEX.match(line)
        if match:
            # Hack to be compatible with IpAddrParser
            self.interface['link'] = 'ether'
            addr = "%s/%s" % (match.group(1), match.group(2))
            try:
                self.interface['addr'].append(addr)
            except KeyError:
                self.interface['addr'] = [addr]
        return self.parseParameters

    def stop(self):
        self.saveInterface()

class RouteParser(TextParser):
    IP_REGEX = r"[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}"
    ROUTE_REGEX = re.compile(
        # ip ip ip
        r"(?P<dest>%s)\s+(?P<gateway>%s)\s+(?P<genmask>%s)\s+" % (IP_REGEX, IP_REGEX,  IP_REGEX)
        # UG 1000 0 0 eth0
        + r"\S+\s+[0-9]+\s+[0-9]+\s+[0-9]+\s+(?P<interface>\S+)")

    def __init__(self, input, filename=None):
        self.routes = []
        TextParser.__init__(self, input, self.parseRoute, filename)

    def parseRoute(self, line):
        match = self.ROUTE_REGEX.match(line)
        if match:
            self.routes.append(match.groupdict())
        return self.parseRoute

def runCommand(cmd_args):
    try:
        # Run with empty environment to avoid translation to user locale
        process = Popen(cmd_args, stdout=PIPE, env={})
        stdout, process_stderr = process.communicate()
        if process.returncode:
            print >>stderr, 'Command "%s" fails with code %s' % (
                " ".join(cmd_args), process.returncode)
            exit(1)
        return stdout
    except OSError, err:
        print >>stderr, 'Unable to run command "%s": %s' \
            % (" ".join(cmd_args), err)
        exit(1)

def isLoopbackAddress(interface, address):
    ip = IPmakenet(address)
    if ip in LOOPBACK:
        return True
    if interface['name'] == "lo":
        return True
    return False

def removeLoopbackAddresses(interface):
    interface['addr'] = filter(lambda addr: not isLoopbackAddress(interface, addr), interface['addr'])

def getInterfaces(ifconfig_format, ifconfig=None):
    # Run ifconfig -a and parse output
    if not ifconfig:
        ifconfig = runCommand(IPADDR)
        ifconfig = StringIO(ifconfig)
    if ifconfig_format == 'ifconfig':
        parser = IfconfigParser(ifconfig, '"%s"' % " ".join(IFCONFIG))
    else:
        parser = IpAddrParser(ifconfig, '"%s"' % " ".join(IPADDR))

    # Store interfaces list, but skip loopback (lo)
    interfaces = {}
    for interface in parser.interfaces:
        if interface['link'] != 'ether':
            print >>stderr, "Info: Skip interface %s (type %s)" % (
                interface['name'], interface['link'])
            continue
        removeLoopbackAddresses(interface)
        if not interface['addr']:
            continue
        interfaces[interface['name']] = interface
    return interfaces

def getRoutes(routes=None):
    if not routes:
        routes = runCommand(ROUTE)
        routes = StringIO(routes)
    parser = RouteParser(routes, '"%s"' % " ".join(ROUTE))
    return parser.routes

def IPmakenet(ip_mask):
    try:
        # Try IPy version 0.54+ constructor
        return IP(ip_mask, make_net=True)
    except TypeError:
        pass

    # Fallback for older version of IPy
    ip, mask = ip_mask.split('/', 1)
    if '.' in mask:
        mask_str = mask
        mask = IP(mask).int()
    else:
        mask_str = str(mask)
        mask = 2**32 - (2**(32 - int(mask)))
    ip = str(IP( IP(ip).int() & mask ))
    ip = IP('%s/%s' % (ip, mask_str))
    return ip

def createDesc(interfaces, routes, version=None):
    interface_id = 0
    network_id = 0

    desc = Desc(version)
    firewall = Firewall(FIREWALL_QUEUE, FIREWALL_NAME, id=1, type="nufw")
    desc.firewalls.append(firewall)
    for interface_dict in interfaces.itervalues():
        interface_id += 1
        interface = Interface(firewall, interface_dict["name"], id=interface_id)
        address_id = 1
        for address in interface_dict['addr']:
            ipv4 = IP(address.split('/', 1)[0])
            addr = Address(ipv4, id=address_id)
            interface.append(addr)
            address_id += 1
        firewall.interfaces.append(interface)

        net_name = interface_dict["name"].upper()
        subnet_index = 0
        for address in interface_dict['addr']:
            ip_mask = IPmakenet(address)
            network_id += 1
            if subnet_index:
                name = "%sALIAS%s" % (net_name, subnet_index)
            else:
                name = net_name
            network = Network(name, "ipv4", ip_mask, id=network_id)
            if not subnet_index and "gateway" in interface_dict:
                gateway = IP(interface_dict["gateway"])
                InternetConnection(network, interface, gateway)
            else:
                DirectConnection(network, interface)
            desc.networks.append(network)
            subnet_index += 1

    for route in routes:
        dest = IPmakenet('%s/%s' % (route['dest'], route['genmask']))
        gateway = IP(route['gateway'])
        if dest == INTERNET or gateway == IP('0.0.0.0'):
            continue

        try:
            interface = searchFirst(lambda item: item.name == route['interface'], firewall.interfaces)
        except KeyError:
            print "Info: Skip route %s/%s=>%s (no interface %r)" % (
                route['dest'],
                route['genmask'],
                route['gateway'],
                route['interface'])
            continue
        try:
            network = searchFirst(lambda item: item.addr == dest, desc.networks)
        except KeyError:
            network_id = len(desc.networks) + 1
            name = "NET%s" % network_id
            network = Network(name, "ipv4", dest, id=network_id)
            desc.networks.append(network)
        RoutedConnection(network, interface, IP(gateway))

    return desc

def writeDescXML(desc_version, output, filename, ifconfig_format, ifconfig=None, routes=None):
    interfaces = getInterfaces(ifconfig_format, ifconfig)
    routes = getRoutes(routes)
    for route in routes:
        dest = IPmakenet('%s/%s' % (route['dest'], route['genmask']))
        if dest != INTERNET:
            continue
        interface = interfaces[route['interface']]
        interface['gateway'] = route['gateway']
    desc = createDesc(interfaces, routes, desc_version)
    dump(desc, output)

